{{
********************************************
* jvmVpCore - Virtual Peripheral functions *
********************************************
}}


{{

This is a Work In Progress (WIP).

This is not a request for coding critique. Suggestions for real improvements
rather than addressing subjective cosmetic cruft would be appreciated.

Hi, If you're reading this, you've downloaded a work that may or may not be
included in the Propeller JVM. I have had some experiences with assembly over
my 20 year software engineering career; they were mostly in "support of" large
embedded systems and multiprocessor designs to get C/C++ booted, for gdb, or
for ISRs handling.

My successful career leads me to believe I can cross any river or ocean, and
thus became involved with this after toying with propeller for some projects.

This is my first PASM attempt. It is not optimized, nor is it complete.
The TIMER, DAC, and PWM work; the ADC and UART devices are to be implemented.
I find it unlikely that all of this can happen in just one COG in 8.68us 8-)
Thus, this framework if possible may need replication to other COGS or be
done in a different way. It is expected that 2 instances of this object will
be required to implement full features of 6 virtual peripherals.

This code is not currently added to JVM build. I am encouraged to make it
public since it contains requirements documentation and a general framework
that may be included in the JVM. I have tested a variant of Timer with the
JVM. I will add this to JVM later pending some issue resolution.

All timings are verified on an O'scope.
Performance summary:

Device - Budget - Startup - R/W Update - Overhead - Device Time - Dev Count - Total - Headroom - Dev Budget
 Timer   8,680      320       1,400        220          400           6       5,440     3,240        940 
 Timer   8,680      320       1,400        220          400           3       3,580     5,100      2,100 
 DAC     8,680      320       1,400        220          620           6       6,760     1,920        940 
 DAC     8,680      320       1,400        220          620           3       4,240     4,440      2,100 
 PWM     8,680      320       1,400        220          980           6       8,920     (240)        940 
 PWM     8,680      320       1,400        220          980           3       5,320     3,360      2,100 

}}

{{
/**
 *
 * Implementing Virtual Peripherals (VP) requires a process to run in the background.
 * This is a good opportunity for Propeller, although it takes another COG out of
 * the user's hands. We don't have a strategy for the user to control a cog yet.
 * One possibility is extending the VP set to more cogs. Another is to provide a
 * way for a user binary to execute on a cog. These are TBD of course.
 *
 * The elements that need a background process are: ADC, DAC, PWM, Timer, and UART.
 * It is expected that one Propeller COG can behave as the VP thread which does
 * processing ever 8.68us. The goal is to have all VP's be processed in that time*.
 *
 * *Note: It is very likely that more than one COG should be used to realize the full
 * potential of the VP design. Having cog/hub shared memory access is very expensive
 * and may not allow time for even the best asm programmer to do everything in 8.68us.
 * Using a time-multiplex design would make all features possible, but this would reduce
 * the granularity of events.
 *
 * One can argue that certain functions such as the ADC do not need a timer process,
 * but the conversion can be made more accurate if done in time slices.  The DAC
 * must be done as part of a timer process since it is more or less just a PWM.
 *
 * Common VP design requirements:
 *
 * - Up to 6 VP should run at once.
 * - All VP elements should run every 8.68us.
 * - VP pin usage is mutually exclusive.
 *
 * - Requirement is to have up to 6 PWM or 6 ADC or 6 DAC or 6 Uart run at a time.
 *   The java code clearly uses only 1 timer at any given moment, the VP of the
 *   timer can be assigned and would replace one of the other VPs in the list above.
 *
 * - jvmNative method nm_core_CPU_uninstall stops a VP this is a good place to
 *   arrange notification for the cog to not update elements.
 *
 * Design requirements for ADC:
 *
 * - ADC will be initialized by native call.
 * - ADC value should be readable by user. Variable for this written by COG/read by java.
 * - ADC pins should be selectable one for input one for reference output.
 * - External RC network is required on the input pin for ADC to function.
 * - Input measurements are limited from 0 to 3.3VDC.
 * - Up to 6 ADC can be used at once; the code will update measurments on each 8.68us tick.
 *
 * Design requirements for DAC:
 *
 * - DAC will be initialized by native call.
 * - DAC should be writeable by user. Variable for this written by java, read by COG.
 * - DAC output pins should be selectable.
 * - External RC network is required on the output pin for DDC to function.
 * - Output measurements are limited from 0 to 3.3VDC.
 * - Up to 6 DAC can be used at once; the code will update output on each 8.68us tick.
 *
 * Design requirements for PWM:
 *
 * - PWM pins should be selectable.
 * - PWM will be initialized by native call.
 * - Each PWM pin's pulse cycle characteristics can be updated any time.
 * - Up to 6 PWM can be used at once; the code will update output on each 8.68us tick.
 *
 * Design requirements for Timer:
 *
 * - Timer should tick every 8.68us.
 * - Timer should start when initialzied by native call.
 * - Timer value should latch on demand from native call.
 * - The Propeller COG process main loop with wait is the basis for the timer.
 * - Timer variables will be incremented by COG and read by java when running.
 * - Timer latch message can be sent from java to COG and cleared by COG for java read.
 *
 * Design requirements for Uart:
 *
 * - Uart will operate in one direction per instance.
 * - Uart will buffer input or output at 256 bytes each.
 * - Uart receiver using HW flow control will stall sender when <= 16 RX buffer bytes are free.
 * - Uart receiver using HW flow control will remove back-pressure when > 16 bytes are free.
 * - Uart baud rate will be selectable from 600 to 57600 BPS.
 * - Uart stop bits will be selectable from 1 to 8.
 * - Uart level sense can be inverted for both data bits and control lines.
 * 
 * - Up to 6 uni-directional Uarts will be allowed.
 * 
 */

/**
 *
 * COG Initialization and Start/Stop design
 *
 * The jvmVpCore start method should be called at system startup (jd.init would be good).
 * Once started, the cog should run freely although stop is possible since all
 * initialization required for cog function should be put into start.
 *
 * VP Start/Stop. VPs are started by the given init or install method below.
 * Calls to jvmVpCore vp.startVP(addr, bank, type) can be added to these methods.
 *
 * nm_core_Uart_txInit(bank)
 * nm_core_Uart_rxInit(bank)
 * nm_core_Timer_init(bank)
 * nm_core_PWM_install(bank)
 * nm_core_DAC_install(bank)
 * nm_core_ADC_install(bank)
 *
 * All VPs are stopped individually by the nm_core_CPU_uninstall method.
 * Call to jvmVpCore vp.stopVP(bank) has been added.
 *
 */
 
/**
 * Timer design -- implemented. Timer can be in any VP slot.
 *
 * This is the easiest design as it involves "generally" incrementing the counter
 * on every pass of the waitcnt timeout in the main cog loop. More detail follows.
 *
 * Timer should start when initialzied by native call. At start time the nmTimer
 * registers will be cleared by java code. This init call is our opportunity to
 * know that the timer is reset, and reset our internal timer accordingly.
 *
 * The nmTimer register is written by the JVM/Java before startup. Afterwards
 * only the COG should write the register and JVM/Java read only the register
 * until the interface is disabled. When disabled, the COG will not access any register.
 * The nmLatchTimer register is written by cog, and read only by JVM/Java.
 * The nmLatchItMsg is set by JVM/Java to signal a timer latch; once written this
 * this value should not be written again until after it is read as clear by JVM/Java. 
 *
 * To prevent a deadlocks the COG and JVM/Java should obey the access rules.
 *
 * The cog will clear its internal timer on init and update latches on each tick.
 *
 * ' file stamp\core\Timer.java (offsets within a VP bank) 
 * nmPeriphISRVec = $00  '// offset for all VPs indicates the bank number VPTYPE*0x10
 *
 * This nmLatchTimer register set is not touched by cog according to jvmNative use.
 *
 * nmLatchTimer1  = $08  '// LSB of 32 bit timer latch
 * nmLatchTimer2  = $09  '
 * nmLatchTimer3  = $0A  '
 * nmLatchTimer4  = $0B  '// MSB of 32 bit timer latch ... not touched by cog
 *
 * This nmTimer register set is written once by Java/JVM at startup and accessed
 * as read or write by the process. Shitty design alert. At least java start
 * writes the counter before starting the timer. This is Little Endian like Prop.
 * 
 * nmTimer1       = $04  '// LSB of 32 bit timer ... read by cog / write by JVM
 * nmTimer2       = $05  '
 * nmTimer3       = $06  '
 * nmTimer4       = $07  '// MSB of 32 bit timer ... read by cog / write by JVM
 *
 * nmLatchItMsg   = $02
 * 
 * '///////// these registers are redefined to allow long access for Timer //////////
 */
 
/**
 * PWM design -- implemented. 6 PWM / COG.
 *
 * Only the initial state is known to not work at this time; it may be added later.
 *
 * The PWM is implemented according to the following rules:
 *
 * @param highTime the number of 8.68us the waveform should be high.
 * @param lowTime the number of 8.68us the waveform should be low.
 * @param highZero Boolean value (always true) that tells wether highTime != 0.
 * @param lowZero Boolean value (always true) that tells wether lowTime != 0.
 * The zero counter is niether set nor read by the COG.
 *
 * The registers below are read-only by COG and written by JVM/Java:
 *
 * 'file stamp\core\PWM.java (offsets within a VP bank) 
 * nmPWMPort     = $0E  '// 0 for PortA, 1 for PortB (reserved: 2 for PortC, 3 for PortD)
 * nmPWMPin      = $0F  '// bitmask (1 of 8 set)
 *
 * nmPWMhighTime = $0C
 * nmPWMlowTime  = $0D
 *
 * nmPWMhighZero, nmPWMlowZero, count, and state are ignored
 *
 * ///////// these registers are redefined to allow long access for PWM //////////
 *
 */
 
/**
 * ADC Design -- not implemented
 *
 * nmAdcValue    = $0b  '// current ADC value
 * nmInPort      = $0c  '// 0 for port0, 1 for port1
 * nmInPin       = $0d  '// bitmask (1 of 8 set)
 * nmOutPort     = $0e  '// 0 for port0, 1 for port1
 * nmOutPin      = $0f  '// bitmask (1 of 8 set)   
 */
 
/**
 * DAC Design -- implemented. Needs method optimization
 *
 * nmDacValue    = $0b  '// current ADC value
 * nmDacPort     = $0e  '// 0 for port0, 1 for port1
 * nmDacPin      = $0f  '// bitmask (1 of 8 set)   
 */
 
/**
 * UART Design -- not implemented.
 *
 * 'file stamp\core\Uart.java (offsets within a VP bank) 
 * nmUartDataH    = $02  '// transmit register highbyte
 * nmUartDataL    = $03  '// transmit register lowbyte 
 * nmUartInvert   = $05  '// 0 for normal, 1 for invert
 * nmUartBaud     = $06  '// 1000000/(bps * 4.34)
 * nmUartStopBits = $07  '// Tx only: 00000001 for 1 stop bit, 00000011 for 2 etc
 * nmUartPort     = $0C  '// 0 for PortA, 1 for PortB (reserved: 2 for PortC, 3 for PortD)
 * nmUartPin      = $0D  '// bit mask (1 of 8 set)
 * nmUartHSPort   = $0E  '// 0 for PortA, 1 for PortB (reserved: 2 for PortC, 3 for PortD)
 * nmUartHSPin    = $0F  '// bitmask (1 of 8 set)
 * nmUartBufferH  = $0A  '// address of 256 byte buffer (aligned to $XX00) 
 * nmUartBufferL  = $0B
 * nmUartHead     = $08  '//Index of the next free byte in the buffer. 
 * nmUartTail     = $09  '//Index of the oldest byte in the buffer.
 *                       '//If tailPointer == headPointer the buffer is empty.
 * 'Uart baud constants
 * speed600   = 192  'The speed value for 600 baud.
 * speed1200  = 96   'The speed value for 1200 baud.
 * speed2400  = 48   'The speed value for 2400 baud.
 * speed4800  = 24   'The speed value for 4800 baud.
 * speed7200  = 16   'The speed value for 7200 baud.
 * speed9600  = 12   'The speed value for 9600 baud.
 * speed14400 = 8    'The speed value for 14400 baud.
 * speed19200 = 6    'The speed value for 19200 baud.
 * speed28800 = 4    'The speed value for 28800 baud.
 * speed38400 = 3    'The speed value for 38400 baud.
 * speed57600 = 2    'The speed value for 57600 baud.
 */
  
}}


CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000
  '_clkmode = xtal1 + pll8x
  '_xinfreq = 10_000_000

  VP_TICKCLK  = 667   'this value with waitcount makes 8.68us tick period
  VP_TICKMUL  = 10_000_000
  VP_DEBUG_PIN = 1
{{
/**
 * @enum VPTypeEnum
 * VP_NONE=0,  VP_TIMER, VP_ADC, VP_DAC, VP_PWM, VP_UART_RX, VP_UART_TX
 */
 _ typedef enum VPTypeEnum_t
}}
  #0, VP_NONE, VP_TIMER, VP_ADC, VP_DAC, VP_PWM, VP_UART_RX, VP_UART_TX

CON
{ jump table variants - jump table has problems - delete
  WRITE         = $0100_0000 'write command
  READ          = $0200_0000 'read command
  ABSWRITE      = $0300_0000 'absolute addr write command
  ABSREAD       = $0400_0000 'absolute addr read command
  COMMAND       = $0500_0000 'command
}
{ bit switch variants }
  WRITE         = $0100_0000 'write command
  READ          = $0200_0000 'read command
  ABSWRITE      = $0400_0000 'absolute addr write command
  ABSREAD       = $0800_0000 'absolute addr read command
  COMMAND       = $1000_0000 'command

  RESPONSE      = $00f0_0000 'to OR with return value to vpCommand
  ANYCMD        = WRITE | READ | ABSREAD | ABSWRITE | COMMAND
  ANYCMD_SHFT   = 24
  REGADR        = $0000_01ff ' one of 512 registers


CON ' COG pin control addresses
' It is not possible to set OUTA from one COG to another.
' This explains why DIRA[pin]~~ in spin init code would not work.
' Using these addresses and readAbs/writeAbs methods we can set values.
  INA_ADDR      = $1F2
  OUTA_ADDR     = $1F4
  DIRA_ADDR     = $1F6
  CTRA_ADDR     = $1F8
  FRQA_ADDR     = $1FA
  PHSA_ADDR     = $1FC
  VCFG_ADDR     = $1FE
  VSCL_ADDR     = $1FF

CON  
  SDEBUG        = 0
      
OBJ
  def: "jvmDefines"
  da:  "jvmVpData"
  sio: "jvmSerial"
  dbg: "PASDebug" '<---- Added for Debugger 

PUB main | addr, pw

  if SDEBUG
    sio.start(31,30,0,9600)
    sio.rx ' uncomment to pause before demo

  'dbg.start(31,30,@entry)
{
  startDAC(def#nmPeriph0Bank, 0, 1, 0)
  startDAC(def#nmPeriph1Bank, 0, 2, 25)
  startDAC(def#nmPeriph2Bank, 0, 3, 128)
  startDAC(def#nmPeriph3Bank, 0, 4, 220)
  startDAC(def#nmPeriph4Bank, 0, 5, 254)
  startDAC(def#nmPeriph5Bank, 0, 6, 1)
}  
  pw := 1

  startPWM(def#nmPeriph0Bank, 0, 1, pw, pw)
  startPWM(def#nmPeriph1Bank, 0, 2, pw, pw)
  startPWM(def#nmPeriph2Bank, 0, 3, pw, pw)
  startPWM(def#nmPeriph3Bank, 0, 4, pw, pw)
  startPWM(def#nmPeriph4Bank, 0, 5, pw, pw)
  'startPWM(def#nmPeriph5Bank, 0, 6, pw, pw)
{
  startPWM(def#nmPeriph0Bank, 0, 1, pw, pw)
  startPWM(def#nmPeriph1Bank, 0, 2, pw+1, pw)
  startPWM(def#nmPeriph2Bank, 0, 3, pw, pw+1)
  startPWM(def#nmPeriph3Bank, 0, 4, pw+2, pw)
  startPWM(def#nmPeriph4Bank, 0, 5, pw, pw+2)
  startPWM(def#nmPeriph5Bank, 0, 6, pw+1, pw+1)
}
{
  startVP(def#nmPeriph0Bank, VP_TIMER)
  startVP(def#nmPeriph1Bank, VP_TIMER)
  startVP(def#nmPeriph2Bank, VP_TIMER)
  startVP(def#nmPeriph3Bank, VP_TIMER)
  startVP(def#nmPeriph4Bank, VP_TIMER)
  startVP(def#nmPeriph5Bank, VP_TIMER)
}
  'start

  'repeat  
  addr := def#nmPeriph0Bank

  ' use repeat reg access during test to ensure we have room for user service    
  if SDEBUG
    sio.tx($d)
    writeReg(addr+def#nmLatchTimer1, readReg(addr+def#nmTimer1))
    sio.hex(readReg(addr+def#nmTimer1),8)
    sio.tx(" ")
    sio.hex(readReg(addr+def#nmLatchTimer1),8)
    pause(200)
    repeat
      readPWM(def#nmPeriph0Bank)
      readPWM(def#nmPeriph1Bank)
      readPWM(def#nmPeriph2Bank)
      readPWM(def#nmPeriph3Bank)
      readPWM(def#nmPeriph4Bank)
      readPWM(def#nmPeriph5Bank)
      pause(10000)
  else
    repeat
      writeReg(addr+def#nmLatchTimer1, readReg(addr+def#nmTimer1))
      'readReg(addr+def#nmTimer1)
      'writeReg(addr+def#nmLatchTimer1, 0)
    
PUB start
  if cog <> 0
    return ' dont start more than once
  cog := cognew(@entry, @vpCommand) + 1
    
PUB stop
  if cog
    cogstop(cog~ - 1)
  cog := 0

PUB writeReg(address,value) | saddr
{{ write value to vpRambank register in VP cog }}
  saddr := address
  address -= def#nmPeriph0Bank
  if address > def#MAX_COG_VPS*16+1
    return  ' don't go out of bounds
  address := RESPONSE | WRITE | address
  vpData  := value
  vpCommand := address
  repeat while vpCommand <> 0
  
  if SDEBUG
    sio.str(string($d,"WRR "))
    sio.hex(address,8)
    sio.tx(" ")
    sio.hex(value,4)
    readReg(saddr)
     
PUB readReg(address): rc | saddr
{{ read value from vpRambank register in VP cog }}
  saddr := address
  address -= def#nmPeriph0Bank
  if address > def#MAX_COG_VPS*16+1
    return  ' don't go out of bounds
  address := RESPONSE | READ  | address
  vpCommand := address
  repeat while vpCommand <> 0
  result := vpData

  if SDEBUG
    sio.str(string($d,"RDR "))
    sio.hex(address,8)
    sio.tx(" ")
    sio.hex(result,4)
     
PUB writeAbs(address,value) | saddr
{{ write value to vpRambank register in VP cog }}
  saddr := address
  address := RESPONSE | ABSWRITE | address
  vpData  := value
  vpCommand := address
  repeat while vpCommand <> 0
  
  if SDEBUG
    sio.str(string($d,"WRA "))
    sio.hex(address,8)
    sio.tx(" ")
    sio.hex(value,4)
    readAbs(saddr)
     
PUB readAbs(address): rc | saddr
{{ read value from vpRambank register in VP cog }}
  saddr := address
  address := RESPONSE | ABSREAD  | address
  vpCommand := address
  repeat while vpCommand <> 0
  result := vpData

  if SDEBUG
    sio.str(string($d,"RDA "))
    sio.hex(address,8)
    sio.tx(" ")
    sio.hex(result,4)
    
PUB enableVp(address,value) | saddr
{{ tell cog to use address as jump vector. TBD value = 0 may disable later. }}
  saddr := address
  address -= def#nmPeriph0Bank
  address := RESPONSE | COMMAND  | address
  vpData  := value
  vpCommand := address
  repeat while vpCommand <> 0
  
  if SDEBUG
    sio.str(string($d,"ENA "))
    sio.hex(address,8)
    sio.tx(" ")
    sio.hex(value,4)
    readReg(saddr)
  
  
PUB pause(ms)
  if ms < 10
    result := ((clkfreq / 1000000) * 9050) / 10
  elseif ms < 100    
    result := ((clkfreq / 1000000) * 9950) / 10
  elseif ms < 1000 
    result := ((clkfreq / 1000000) * 10090) / 10
  else
    result := (clkfreq / 1000)
  result *= ms
  waitcnt(cnt+result)
    

PUB startDAC(bank, port, pin, value) | pinbit, tmp, rback
{{
/**
 * @fn startPWM
 * Call from native init/install method to start
 * @param bank - bank number to start
 * @param port - 0..3
 * @param pin  - 0..7
 * @param value - number of 8.68us ticks to keep pulse high
 * @note - port will always be zero. pin will be saved as 2^n pin bit number. 
 */
 _ void startVP(int bank, int port, int pin, int hcnt, int lcnt);
}}
  start ' make sure vp is started before sending commands

  pin += port*8
  pinbit := 1 << pin

  ' we have to tell the VP COG to set these values
  tmp := readAbs(DIRA_ADDR)
  tmp |= pinbit
  writeAbs(DIRA_ADDR,tmp)
  tmp := readAbs(DIRA_ADDR)
  rback := tmp & pinbit

  tmp := readAbs(OUTA_ADDR)
  tmp |= pinbit         'start high. first transaction toggles low
  writeAbs(OUTA_ADDR,tmp)

  writeReg(bank + def#nmDacPin, pinbit)
  writeReg(bank + def#nmDacValue, value)

  startVP(bank, VP_DAC)

PUB readDAC(bank)
{{
/**
 * @fn dumpPWM
 * Dump bank PWM register
 * @param bank - bank number to start
 */
 _ void readPWM(int bank);
}}
  start ' make sure vp is started before sending commands
  sio.str(string($d,"Read DAC "))
  sio.hex(bank, 4)
  sio.tx($d)
  readReg(bank)
  readAbs(DIRA_ADDR)
  readAbs(OUTA_ADDR)
  readReg(bank + def#nmDACPin)
  readReg(bank + def#nmDACValue)
  readReg(bank + def#nmDACAccu)

PUB startPWM(bank, port, pin, hcnt, lcnt) | pinbit, tmp, rback
{{
/**
 * @fn startPWM
 * Call from native init/install method to start
 * @param bank - bank number to start
 * @param port - 0..3
 * @param pin  - 0..7
 * @param hcnt - number of 8.68us ticks to keep pulse high
 * @param lcnt - number of 8.68us ticks to keep pulse low
 * @note - port will always be zero. pin will be saved as 2^n pin bit number. 
 */
 _ void startVP(int bank, int port, int pin, int hcnt, int lcnt);
}}
  start ' make sure vp is started before sending commands

  pin += port*8
  pinbit := 1 << pin

  ' we have to tell the VP COG to set these values
  tmp := readAbs(DIRA_ADDR)
  tmp |= pinbit
  writeAbs(DIRA_ADDR,tmp)
  tmp := readAbs(DIRA_ADDR)
  rback := tmp & pinbit
   
  tmp := readAbs(OUTA_ADDR)
  tmp |= pinbit         'start high. first transaction toggles low
  writeAbs(OUTA_ADDR,tmp)
  'used
  writeReg(bank + def#nmPWMPin, pinbit)
  writeReg(bank + def#nmPWMCount, 1) ' don't go negative
  writeReg(bank + def#nmPWMlowTime, lcnt)
  writeReg(bank + def#nmPWMhighTime, hcnt)
  'unused
  writeReg(bank + def#nmPWMState, 0)
  writeReg(bank + def#nmPWMPort, port)

  startVP(bank, VP_PWM)

PUB readPWM(bank)
{{
/**
 * @fn dumpPWM
 * Dump bank PWM register
 * @param bank - bank number to start
 */
 _ void readPWM(int bank);
}}
  start ' make sure vp is started before sending commands
  sio.str(string($d,"Read PWM "))
  sio.hex(bank, 4)
  sio.tx($d)
  readReg(bank)
  readAbs(DIRA_ADDR)
  readAbs(OUTA_ADDR)
  readReg(bank + def#nmPWMPin)
  readReg(bank + def#nmPWMlowTime)
  readReg(bank + def#nmPWMhighTime)
  readReg(bank + def#nmPWMState)
  readReg(bank + def#nmPWMCount)


PUB startVP(bank, type)
{{
/**
 * @fn startVP
 * Call from native init/install method to start VP
 * @param bank - bank number to start
 * @param type - type of VPTypeEnum:
 *   VP_TIMER, VP_ADC, VP_DAC, VP_PWM, VP_UART_RX, VP_UART_TX
 * @note starting with type VP_NONE will not start any VP.
 */
 _ void startVP(int bank, VpTypeEnum type);
}}
  start ' make sure vp is started before sending commands
  result := bank
  result >>= 4
  result &= $7 ' apparently bank 0 is 8. Num 7 ($70) is defined as "block", but not used
  if result => $0 and result =< $5
    'writeReg(bank, type)
    enableVp(bank, type)

      
PUB stopVP(bank)
{{
/**
 * @fn stopVP
 * Call nm_core_CPU_uninstall to stop VP.
 * jvm assembler programmer should make an asm version of this.
 * @param bank - bank number to start
 */
 _ void stopVP(int bank);
}}
{{ call nm_core_CPU_uninstall to stop VP }}
  start ' make sure vp is started before sending commands
  result := bank
  bank := bank - def#nmPeriph0Bank
  result >>= 4
  result &= $7 ' apparently bank 0 is 8. Num 7 ($70) is defined as "block", but not used
  if result => $0 and result =< $5
    'writeReg(bank, 0)
    enableVp(bank, 0)

   
PUB getVpRamAddr
  result := @vpRambank

    
DAT
{{
_______________________________________________________________________________
_______________________________________________________________________________

An assembly cog which performs Javelin-like Virtual Peripheral VP functions.
Executes every 8.68 usec if possible. Currently no over-limit protection.
VP_DEBUG_PIN references should be commented or replaced with NOP at some point.
_______________________________________________________________________________
_______________________________________________________________________________
}}
entry           org 0

{{
_______________________________________________________________________________
Main entry for VP process
}}
'  --------- Debugger Kernel add this at Entry (Addr 0) ---------
'   long $34FC1202,$6CE81201,$83C120B,$8BC0E0A,$E87C0E03,$8BC0E0A
'   long $EC7C0E05,$A0BC1207,$5C7C0003,$5C7C0003,$7FFC,$7FF8
'  -------------------------------------------------------------- 
'______________________________________________________________________________
'
                mov     vpCmdPtr,   par
                mov     vpDataPtr,  vpCmdPtr    'setup dataptr for read/write
                add     vpDataPtr,  #4          '
                or      dira, tickout           'set  tickout pin to output
                mov     time, cnt               'record current time
                add     time, period            'establish next period
                or      dira, #VP_DEBUG_PIN     'debug only
tickLoop        xor     outa, tickout           'toggle tick output for frequency measurement
               'call    #vpTickTock             'debug only. lets us stop on a tick                                                           
                call    #vpUpdate
                call    #vpUserCmd
                waitcnt time, period    'we made it in allocated time :P, wait for next tick
                jmp     #tickLoop               'loop to next tick
{{
_______________________________________________________________________________
PRI function: vpTickTock(...)
Ticks a counter that we can use for debug stop. 
}}
vpTickTock      add     tick, #1
                cmp     tick, tock      wc
                if_b    jmp #vpTickTock_ret
                nop
vpTickTock_ret  ret
{{
_______________________________________________________________________________
PRI function: vpUpdate(...)
Does all VP functions in much less than 8.68us ... if we're lucky :)
App. < 5.4us with user cmd updates for 6 timers.
This update relies on the user enableVp command to take the vp number
destined for rambank[bank+0] and convert it into a procedure address.
}}
vpUpdate        mov     vpAsmBank,  #vpRambank      ' reset bank
                movs    tickTable,  vpAsmBank       ' initial type read addr
                mov     vpAsmIndex, #def#MAX_COG_VPS' reset index
tickTable       mov     vpAsmType,  0-0 wz
                xor     outa,       #VP_DEBUG_PIN   'debug only
                if_nz   jmp         vpAsmType       ' don't add debug stuff here.
doVpJumpReturn  xor     outa,       #VP_DEBUG_PIN   'debug only
                add     vpAsmBank,  #$10            ' increment bank pointer
                movs    tickTable,  vpAsmBank       ' new type read addr
                djnz    vpAsmIndex, #tickTable
vpUpdate_ret    ret
                                
{{
_______________________________________________________________________________
PRI function: vpUserCmd(...)
Processes user commands from vpCmdPtr/vpDataPtr 
}}
'______________________________________________________________________________
'
'--------------------------------------------------------        
' get command
vpUserCmd       rdlong  vpAsmCmd,   vpCmdPtr    wz 'read command into vpAsmCmd
                if_z    jmp         #vpUserCmdExit
                or     outa,       #VP_DEBUG_PIN
                rdlong  vpAsmData,  vpDataPtr   'read data into vpAsmData
                mov     vpAsmVpReg, vpAsmCmd
                and     vpAsmVpReg, cmdVPREG    'no shift reg already in 2^0
'--------------------------------------------------------        
' read from any cog memory/write to user
vpUserAbsRead   test    vpAsmCmd,   cmdABSREAD  wz
                if_z    jmp         #vpUserAbsWrite
                movs    vpUserAbsRdRg, vpAsmVpReg  'mod wrlong instr. is not faster
                nop     'pipe filler
vpUserAbsRdRg   mov     vpAsmCmd, 0-0
                wrlong  vpAsmCmd,   vpDataPtr
                jmp     #vpUserDone
'--------------------------------------------------------                                     
' read from user/write to any cog memory                                    
vpUserAbsWrite  test    vpAsmCmd,   cmdABSWRITE wz
                if_z    jmp         #vpUserRead
                movd    vpUserAbsWrRg, vpAsmVpReg
                nop     'pipe filler
vpUserAbsWrRg   mov     0-0, vpAsmData
                jmp     #vpUserDone
'--------------------------------------------------------        
' read from memory/write to user
vpUserRead      test    vpAsmCmd,   cmdREAD     wz
                if_z    jmp         #vpUserWrite
                add     vpAsmVpReg, #vpRambank
                movs    vpUserRdRg, vpAsmVpReg  'mod wrlong instr. is not faster
                nop     'pipe filler
vpUserRdRg      mov     vpAsmCmd, 0-0
                wrlong  vpAsmCmd,   vpDataPtr
                jmp     #vpUserDone
'--------------------------------------------------------
'read from user/write to vpRambank memory                                     
'upper bound spillover - should never happen
'writeReg limits address ... exits if address out of range.
vpUserWrite     test    vpAsmCmd,   cmdWRITE    wz
                if_z    jmp         #vpUserCommand
                add     vpAsmVpReg, #vpRambank
                movd    vpUserWrRg, vpAsmVpReg
                nop
vpUserWrRg      mov     0-0, vpAsmData
'______________________________________________________________________________
'Command to set jump vector at vpRambank bank# base
'Put jump table method address in vprambank block bank offset 0
'if vpreg = rambank % $10. Reduces time for running update loop.
'Update loop goes from double indirection to single indirection.
vpUserCommand   test    vpAsmCmd,   cmdCOMMAND    wz
                if_z    jmp         #vpUserDone
                add     vpAsmVpReg, #vpRambank      'add rambank address
                movd    vpUserCmnd0,vpAsmVpReg      'setup to write VP index at bank# base
                movs    userTable,  vpAsmVpReg      'setup to read from user table index
vpUserCmnd0     mov     0-0,vpAsmData               'get VP index from data and write to bank# base
                mov     vpAsmType,  #vpCoreJumpTable'user jump table
userTable       add     vpAsmType,  0-0 wz          ' + index
                movd    userJmpJmp, vpAsmVpReg      'modify jump instruction address location
                movs    userJmpJmp, vpAsmType       'modify jump instruction address
                nop     'nothing to do aaarrggh
userJmpJmp      mov     0-0,        0-0
' last item in list - no jmp #vpUserDone here
'--------------------------------------------------------
vpUserDone      mov     vpAsmCmd,   #0
                wrlong  vpAsmCmd,   vpCmdPtr
vpUserCmdExit   andn    outa, #VP_DEBUG_PIN
vpUserCmd_ret   ret
{{
_______________________________________________________________________________
_______________________________________________________________________________
}}
{{
_______________________________________________________________________________
PRI function: vpAsmNone(...)
For jump table entries not covered/valid in VP type/slot index 
}}
vpAsmNone       jmp     #doVpJumpReturn      
{{
_______________________________________________________________________________
PRI function: vpAsmTimer(...)
The function of the timer is to increment on every "system" tick.
Increment 32bits timer in bank specified by vpAsmBank
}}
'______________________________________________________________________________
'
'only called when vpType is TIMER
' using local storage and self-modifying add ....
' timer read/clear/latch commands must be used more time to do that now per command tick
' 400ns
vpAsmTimer      mov     tBnk, vpAsmBank
                add     tBnk, #def#nmTimer1
                movd    vpAsmTimeTick, tBnk
                nop     ' fill pipe ... too bad we need it.      
vpAsmTimeTick   add     0-0, #1
                jmp     #doVpJumpReturn      
'______________________________________________________________________________
'
{{
_______________________________________________________________________________
PRI function: vpAsmAdc(...)
deliver ADC value

}}
vpAsmAdc        jmp     #doVpJumpReturn      
{{
_______________________________________________________________________________
PRI function: vpAsmDac(...)
Produce DAC output
_______________________________________________________________________________
A DAC is nothing more than a PWM with a "duty cycle". The value is created by
the proportion of high/low time. Thus, an output voltage of 3.3 is obtained when
the duty cycle is near or at 100% high. And 0 is when the duty cycle is near or
at 100% low. With a 50% duty cycle the output voltage would be 1.65. This all
depends on the RC network of course. The R*C time constant should be near
255*8.68us or 2.2ms.
nmDacValue 
nmDacPort  
nmDacPin   
nmDacAccu  
}}
'______________________________________________________________________________
'Load values from relative addresses using  diffs.
'///////// This could use some optimization as in PWM below. /////////////
vpAsmDac        mov     tBnk,   vpAsmBank        'target bank
                add     tBnk,   #def#nmDacAccu  'add for count source
                movs    vpAsmDacAccRd,  tBnk    'set accu source
                movd    vpAsmDacAccWr,  tBnk    'set accu dest
                add     tBnk,   vpAsmDacAVAdiff
vpAsmDacAccRd   mov     dacAccu,    0-0 wz      'get accum
                if_nz   jmp     #vpAsmDacSave           'if count non-zero keep state 'til next time
                movs    vpAsmDacValRd,  tBnk    'set value source
                add     tBnk,   vpAsmDacAPVdiff
                movs    vpAsmDacPinRd,  tBnk    'set pin source  
vpAsmDacValRd   mov     dacValue,   0-0         'get value
vpAsmDacPinRd   mov     dacPinBit,  0-0         'get pin bits
'--------------------------------------------------------        
                xor     outa,       dacPinBit
                test    outa,       dacPinBit   wz ' MUST test this bit
                if_z    jmp     #vpAsmDacLoadLo
                mov     dacAccu,    dacValue    ' high count = value
                add     dacAccu,    #1          ' fudge. can't make timing w/more insts. to accomodate zero
                jmp     #vpAsmDacSave
vpAsmDacLoadLo  mov     dacAccu,    #256        ' low count  = 256 - value
                sub     dacAccu,    dacValue
'--------------------------------------------------------        
vpAsmDacSave    sub     dacAccu, #1             'decr accum
vpAsmDacAccWr   mov     0-0,    dacAccu         'set accum
'--------------------------------------------------------        
vpAsmDacJmp     jmp     #doVpJumpReturn
'______________________________________________________________________________
'
{{
_______________________________________________________________________________
PRI function: vpAsmPwm(...)
Produce PWM output
PWM high/low is controlled by "state" ...
    byte[vpRambank + bank + def#nmPWMHtime + state] ' state = 0 for hi time, 1 for low time
}}
'______________________________________________________________________________
'
'load values from relative addresses using  diffs
'1.25us .... appears to be upper end for 6 VPs
'______________________________________________________________________________
'
vpAsmPwm        mov     tBnk,   vpAsmBank       'get bank addr
                add     tBnk,   #def#nmPWMCount 'add for count source
                movd    vpAsmPwmCount,  tBnk    'set count dest for sub
                movd    vpAsmPwmCtWr,   tBnk    'set count adr
vpAsmPwmCount   cmpsub  0-0,    #1  wz          'comp to 1 if dst > src, subtract 1
                if_nz   jmp     #vpAsmPwmJmp    'if count negative, reload with new PW
                add     tBnk,   vpAsmPwmAPCdiff 'add to count for pin source
                movs    vpAsmPwmPin,    tBnk    'set pin source now
                movs    vpAsmPwmXor,    tBnk    'set pin source now
                sub     tBnk,   vpAsmPwmAPLdiff 'go back to low time addr
vpAsmPwmPin     test    outa,   0-0  wz         'MUST TEST the bit ... xor wz wont do it
                if_z    sub     tBnk,   vpAsmPwmALHdiff 'sub back LH to get high PWM addr
                movs    vpAsmPwmCtWr,   tBnk    'set time source to add back on count
vpAsmPwmXor     xor     outa,   0-0             'no need to cheat now. bigger budget.
vpAsmPwmCtWr    add     0-0,    0-0             'add PWM to count
vpAsmPwmJmp     jmp     #doVpJumpReturn      
'______________________________________________________________________________
'
{{
_______________________________________________________________________________
PRI function: vpAsmUartTx(...)
produce Uart output
}}
vpAsmUartTx    jmp     #doVpJumpReturn      
      
{{
_______________________________________________________________________________
PRI function: vpAsmUartRx(...)
Deliver Uart receive function
}}
vpAsmUartRx    jmp     #doVpJumpReturn      

{{
_______________________________________________________________________________
_______________________________________________________________________________
DAT ASM Variables
}}

MASK_BYTE       long    $FF
MASK_WORD       long    $FFFF

cmdRESPONSE     long    RESPONSE    'to OR with return value to vpCommand
cmdWRITE        long    WRITE       'write command
cmdREAD         long    READ        'read command
cmdABSWRITE     long    ABSWRITE    'write command
cmdABSREAD      long    ABSREAD     'read command
cmdCOMMAND      long    COMMAND     'command command - just used for enable/disable jump vector
cmdANY          long    ANYCMD
cmdVPREG        long    REGADR

cog             long    0
tickout         long    |< 25     'APIN=0
infreq          long    5_000_000                 'overloaded by user with start(freq)
period          long    VP_TICKCLK/2 * (VP_TICKMUL / _xinfreq) '115200 Hz (8.68S period) (_clkfreq / period)
time            long    0
ttime           long    0

tick            long    0
tock            long    12000 '104

vpRamPtr        long    0         'pointer storage set by .spin
vpAsmIndex      long    0         'the VP index getting an update
vpAsmBank       long    0
vpAsmType       long    0
vpAsmVpReg      long    0

vpAsmCmd        long    0
vpAsmData       long    0
vpCmdPtr        long    0
vpDataPtr       long    0

var1            long    0
var2            long    0
var3            long    0
var4            long    0
tBnk            long    0

vpAsmDacAPVdiff long    def#nmDacPin-def#nmDacValue
vpAsmDacAVAdiff long    def#nmDacValue-def#nmDacAccu

dacAccu         long    0
dacValue        long    0
dacPinBit       long    0

vpAsmPwmAPHdiff long    def#nmPWMPin-def#nmPWMhighTime
vpAsmPwmAPLdiff long    def#nmPWMPin-def#nmPWMlowTime
vpAsmPwmAPCdiff long    def#nmPWMPin-def#nmPWMCount
vpAsmPwmAHCdiff long    def#nmPWMhighTime-def#nmPWMCount
vpAsmPwmALCdiff long    def#nmPWMlowTime-def#nmPWMCount
vpAsmPwmALHdiff long    def#nmPWMlowTime-def#nmPWMhighTime

pwmPinBit       long    0
pwmPnAdr        long    0
pwmState        long    0
pwmCount        long    0
pwmStAdr        long    0
pwmCtAdr        long    0
pwmHiAdr        long    0
pwmLoAdr        long    0

'jump table specifies a range of functions to call. A 0 (zero) entry means do not jump.
'
vpCoreJumpTable long    (@vpAsmNone   - @entry)/4 '0  zero is do nothing
                long    (@vpAsmTimer  - @entry)/4 '1
                long    (@vpAsmAdc    - @entry)/4 '2
                long    (@vpAsmDac    - @entry)/4 '3
                long    (@vpAsmPwm    - @entry)/4 '4
                long    (@vpAsmUartRx - @entry)/4 '5
                long    (@vpAsmUartTx - @entry)/4 '6
                long    (@vpAsmNone   - @entry)/4

vpCommand       long    0 'keep just before rambank
vpData          long    0
vpRambank       long    0 [def#MAX_COG_VPS*16+4] 'registers for Virtual Peripherals (16 bytes per VP), 2 dummy bytes at start

{{
_______________________________________________________________________________
_______________________________________________________________________________
}}
                fit     $1F0